import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.Socket;
import java.security.SecureRandom;
import java.util.Arrays;
import java.util.Hashtable;

import org.bouncycastle.asn1.ASN1Encodable;
import org.bouncycastle.asn1.ASN1ObjectIdentifier;
import org.bouncycastle.asn1.ocsp.BasicOCSPResponse;
import org.bouncycastle.asn1.ocsp.SingleResponse;
import org.bouncycastle.asn1.x509.Extension;
import org.bouncycastle.asn1.x509.Extensions;
import org.bouncycastle.asn1.x509.KeyPurposeId;
import org.bouncycastle.crypto.tls.Certificate;
import org.bouncycastle.crypto.tls.CertificateStatusRequest;
import org.bouncycastle.crypto.tls.CertificateStatusType;
import org.bouncycastle.crypto.tls.DefaultTlsClient;
import org.bouncycastle.crypto.tls.ExtensionType;
import org.bouncycastle.crypto.tls.OCSPStatusRequest;
import org.bouncycastle.crypto.tls.ServerOnlyTlsAuthentication;
import org.bouncycastle.crypto.tls.TlsAuthentication;
import org.bouncycastle.crypto.tls.TlsClient;
import org.bouncycastle.crypto.tls.TlsClientProtocol;
import org.bouncycastle.crypto.tls.TlsExtensionsUtils;


public class OCSP {
	private static final ASN1ObjectIdentifier GOOGLE = new ASN1ObjectIdentifier("1.3.6.1.4.1.11129");
	public static final ASN1ObjectIdentifier X509_SIGNED_CERTIFICATE_TIMESTAMP = GOOGLE.branch("2.4.2");
	public static final ASN1ObjectIdentifier PRECERTIFICATE_POISON = GOOGLE.branch("2.4.3");
	public static final KeyPurposeId PRECERTIFICATE_SIGNING_CERTIFICATE = KeyPurposeId.getInstance(GOOGLE.branch("2.4.4"));
	public static final ASN1ObjectIdentifier OCSP_SIGNED_CERTIFICATE_TIMESTAMP = GOOGLE.branch("2.4.5");
	
	private static Extension getExtension(Extensions extensions, ASN1ObjectIdentifier oid) {
		if (extensions == null) {
			return null;
		}

		return extensions.getExtension(oid);
	}

	
	public static void main(String[] args) throws IOException {
		// TODO Auto-generated method stub
		Socket socket = new Socket("facebook.com", 443);

		TlsClientProtocol tls = new TlsClientProtocol(socket.getInputStream(), socket.getOutputStream(), new SecureRandom()) {
			@Override
			public void connect(TlsClient client) throws IOException {
				allowCertificateStatus = true;
				super.connect(client);
			}
			
			protected void sendClientHelloMessage() throws IOException{
				super.sendClientHelloMessage();
			}

			@Override
			protected void receiveServerHelloMessage(ByteArrayInputStream buf) throws IOException {
				super.receiveServerHelloMessage(buf);

				byte[] sct = (byte[]) serverExtensions.get(ExtensionType.signed_certificate_timestamp);
				if (sct != null) {
					System.out.println("TLS: " + Arrays.toString(sct));
				}
			}

			@Override
			protected void completeHandshake() throws IOException {
				if (certificateStatus != null) {
					BasicOCSPResponse response = BasicOCSPResponse.getInstance(certificateStatus.getOCSPResponse().getResponseBytes().getResponse().getOctets());
					for (ASN1Encodable r : response.getTbsResponseData().getResponses()) {
						Extension extension = getExtension(SingleResponse.getInstance(r).getSingleExtensions(), OCSP_SIGNED_CERTIFICATE_TIMESTAMP);
						if (extension != null) {
							System.out.println("OCSP: " + Arrays.toString(extension.getExtnValue().getOctets()));
						}
					}
				}

				super.completeHandshake();
			}
		};

		TlsClient client = new DefaultTlsClient() {
			@Override
			public Hashtable getClientExtensions() throws IOException {
				CertificateStatusRequest statusRequest = new CertificateStatusRequest(CertificateStatusType.ocsp, new OCSPStatusRequest(null, null));

				@SuppressWarnings("unchecked")
				Hashtable<Integer, byte[]> extensions = TlsExtensionsUtils.ensureExtensionsInitialised(super.getClientExtensions());
				extensions.put(ExtensionType.signed_certificate_timestamp, new byte[0]);
				extensions.put(ExtensionType.status_request, TlsExtensionsUtils.createStatusRequestExtension(statusRequest));
				return extensions;
			}

			@Override
			public TlsAuthentication getAuthentication() throws IOException {
				return new ServerOnlyTlsAuthentication() {
					@Override
					public void notifyServerCertificate(Certificate certificate) throws IOException {
						Extension extension = getExtension(certificate.getCertificateAt(0).getTBSCertificate().getExtensions(), X509_SIGNED_CERTIFICATE_TIMESTAMP);
						if (extension != null) {
							System.out.println("X509: " + Arrays.toString(extension.getExtnValue().getOctets()));
						}
					}
				};
			}
		};

		try {
			tls.connect(client);
		} finally {
			tls.close();
		}
	}

}
