import java.nio.ByteBuffer;
import java.security.cert.CertificateException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Base64;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;

import org.bouncycastle.util.encoders.Base64Encoder;

public class SCTExtension {
	private List<SCT> scts = new ArrayList<SCT>();
	private byte[] data;
	private int lenght = 0;
	private int signedCertificateTimestampListLenght = 0;
	private long decodeTime;
	/**
	 * @return the decodeTime
	 */
	public long getDecodeTime() {
		return decodeTime;
	}

	private static Map<String,String> loggs = new HashMap<>();
	
	/**
	 * @return the loggs
	 */
	public static Map<String, String> getLoggs() {
		return loggs;
	}

	/**
	 * @return the timestampDateString
	 */
	
	public SCTExtension(byte[] data) throws CertificateException
	{
		if(loggs.isEmpty())
		{
			loggs.put("pLkJkLQYWBSHuxOizGdwCjw1mAT5G9+443fNDsgN3BA=","ct.googleapis.com/pilot");
			loggs.put("KTxRllTIOWW6qlD8WAfUt2+/WHopctykwwz05UVH9Hg=","ct.googleapis.com/icarus");
			loggs.put("7ku9t3XOYLrhQmkfq+GeZqMPfl+wctiDAMR7iXqo/cs=","ct.googleapis.com/rocketeer");
			loggs.put("u9nfvB+KcbWTlCOXqpJ7RzhXlQqrUugakJZkNo4e0YU=","ct.googleapis.com/skydiver");
			loggs.put("VhQGmi/XwuzT9eG9RLI+x0Z2ubyZEVzA75SYVdaJ0N0=","ct1.digicert-ct.com/log");
			loggs.put("3esdK3oNT6Ygi4GtgWhwfi6OnQHVXIiNPRHEzbbsvsw=","ct.ws.symantec.com");
			loggs.put("vHjh38X2PGhGSTNNoQ+hXwl5aSAJwIG08/aRfz7ZuKU=","vega.ws.symantec.com");
			loggs.put("rDua7X+pZ0dXFZ5tfVdWcvnZgQCUHpve/+yhMTt1eC0=","ctlog.api.venafi.com");
			loggs.put("QbLcLonmPOSvG6e7Kb9oxt7m+fHMBH4w3/rjs7olkmM=","ctlog.wosign.com");
			loggs.put("pXesnO11SN2PAltnokEInfhuD0duwgPC7L7bGF8oJjg=","ctserver.cnnic.cn");
			loggs.put("NLtq1sPfnAPuqKSZ/3iRSGydXlysktAfe/0bzhnbSO8=","ct.startssl.com");
			loggs.put("iUFEnHB0Lga5/JznsRa6ACSqNtWa9E8CBEBPAPfqhWY=","ct.izenpe.eus");
			loggs.put("yc+JCiEQnGZswXo+0GXJMNDgE1qf66ha8UIQuAckIao=","ct.gdca.com.cn");
			loggs.put("kkow+Qkzb/Q11pk6EKx1osZBco5/wtZZrmGI/61AzgE=","ctlog.gdca.com.cn");
			loggs.put("4BJ2KekEllZOPQFHmESYqkj4rbFmAOt5AqHvmQmQYnM=","www.certificatetransparency.cn/ct");
			loggs.put("AwGd8/2FppqOvR+sxtqbpz5Gl3T+d/V5/FoIuDKMHWs=","ctlog-gen2.api.venafi.com");
			loggs.put("ZcEiNe5l6Bb61JRKt7o0ui0oxZSZBIan6v71fha2T8=","sirius.ws.symantec.com");
			loggs.put("h3W/51l8+IxDmV+9827/Vo1HVjb/SrVgwbTq/16ggw8=","ct2.digicert-ct.com/log");
			loggs.put("aPaY+B9kgr46jO65KB1M/HFRXWeT1ETRCmesu09P+8Q=","ct.googleapis.com/aviator");
			loggs.put("nk/3PcPOIgtpIXyJnkaAdqv414Y21cz8haMadWKLqIs=","ct.wosign.com");
			loggs.put("zbUXm3/BwEb+6jETaj+PAC5hgvr4iW/syLL1tatgSQA=","log.certly.io");
			loggs.put("dGG0oJz7PUHXUVlXWy52SaRFqNJ3CbDMVkpkgrfrQaM=","ct.izenpe.com");
			loggs.put("OTdvVF97Rgf1l0LXaM1dJDe/NHO2U0pINLz3Lmgcg8k=","alpha.ctlogs.org");
		}
		if(data.length == 0)
			throw new CertificateException("Could not parse certificate: java.io.IOException: Empty input");
		
		this.data = data;
		this.lenght = data.length;
		
		this.read();
		
	}
	
	private void read()
	{
		long start = System.nanoTime();
		ByteBuffer listBuf = ByteBuffer.wrap(this.data);
		int listLen = listBuf.getShort() & 0xFFFF;

		while (listBuf.remaining() >= 2) {
			SCT sct = new SCT();
		    int sctLen = listBuf.getShort() & 0xFFFF;

		    byte[] data = new byte[sctLen];
		    listBuf.get(data);

		    ByteBuffer sctBuf = ByteBuffer.wrap(data);

		    sct.version = sctBuf.get() & 0xFF;

		    byte[] logId = new byte[32];
		    sctBuf.get(logId);
		    sct.logId = Base64.getEncoder().encodeToString(logId);

		    long timestamp = sctBuf.getLong();
		    long diff = System.currentTimeMillis() - timestamp;
		    Calendar c= Calendar.getInstance(TimeZone.getTimeZone("UTC"));
		    c.setTimeInMillis(timestamp);
		    
		    SimpleDateFormat simpleDateFormat = 
		    	       new SimpleDateFormat("yyyy-MM-dd HH:mm:ss", Locale.US);
	    	simpleDateFormat.setTimeZone(TimeZone.getTimeZone("UTC"));
	    	sct.timestampDateString = simpleDateFormat.format(c.getTime());
		    

		    int extensionsLen = sctBuf.getShort() & 0xFFFF;
		    byte[] extensions = new byte[extensionsLen];
		    sctBuf.get(extensions);

		    sct.hashAlgorithm = sctBuf.get() & 0xFF;
		    sct.signatureAlgorithm = sctBuf.get() & 0xFF;
		    sct.signatureLen = sctBuf.getShort() & 0xFFFF;
		    byte[] signatureBytes = new byte[sct.signatureLen];
		    sctBuf.get(signatureBytes);
		    sct.signature = Base64.getEncoder().encodeToString(signatureBytes);
		    this.scts.add(sct);
		}
		this.decodeTime = System.nanoTime() - start;
	}


	/**
	 * @return the scts
	 */
	public List<SCT> getScts() {
		return scts;
	}

	/**
	 * @return the lenght
	 */
	public int getLenght() {
		return lenght;
	}


	/**
	 * @return the signedCertificateTimestampListLenght
	 */
	public int getSignedCertificateTimestampListLenght() {
		return signedCertificateTimestampListLenght;
	}

	public class SCT {
		private int lenght = 0;
		private int version = 0;
		private String logId = "";
		private String timestampDateString;
		private int hashAlgorithm;
		private int signatureAlgorithm;
		private int signatureLen;
		private String signature = "";
		
		private SCT()
		{
		}
		
		public String GetLog()
		{
			return loggs.get(this.logId);
		}

		/**
		 * @return the lenght
		 */
		public int getLenght() {
			return lenght;
		}

		/**
		 * @return the version
		 */
		public int getVersion() {
			return version;
		}

		/**
		 * @return the logId
		 */
		public String getLogId() {
			return logId;
		}

		/**
		 * @return the decodeTime
		 */
		public long getDecodeTime() {
			return decodeTime;
		}

		/**
		 * @return the timestampDateString
		 */
		public String getTimestampDateString() {
			return timestampDateString;
		}

		/**
		 * @return the hashAlgorithm
		 */
		public int getHashAlgorithm() {
			return hashAlgorithm;
		}

		/**
		 * @return the signatureAlgorithm
		 */
		public int getSignatureAlgorithm() {
			return signatureAlgorithm;
		}

		/**
		 * @return the signatureLen
		 */
		public int getSignatureLen() {
			return signatureLen;
		}

		/**
		 * @return the signature
		 */
		public String getSignature() {
			return signature;
		}

		
	}
}
