import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.ConnectException;
import java.net.UnknownHostException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.AbstractMap.SimpleEntry;
import java.util.ArrayList;
import java.util.Base64;
import java.util.Base64.Decoder;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map.Entry;

import javax.net.ssl.SSLSocketFactory;
import javax.xml.bind.DatatypeConverter;

import org.apache.commons.io.IOUtils;
import org.bouncycastle.asn1.ASN1InputStream;
import org.bouncycastle.asn1.ASN1ObjectIdentifier;
import org.bouncycastle.asn1.ASN1Sequence;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.x509.CertificatePolicies;
import org.bouncycastle.asn1.x509.Extension;
import org.bouncycastle.asn1.x509.Extensions;
import org.bouncycastle.asn1.x509.PolicyInformation;


public class X509Downloader {
	
	private CertificateFactory certificatefactory;
	private SSLSocketFactory socketFactory;
	private LinkedHashMap<String, Boolean> topSites;
	private String fileName = "data.txt";
	private long handshakeTime = 0;
	private static int siteCount = 0;
	private ArrayList<SimpleEntry<Integer,String>> list;
	private int threads;

	private HashMap<String, String> certTypeMap = new HashMap<>();

	
	private static Extension getExtension(Extensions extensions, ASN1ObjectIdentifier oid) {
		if (extensions == null) {
			return null;
		}

		return extensions.getExtension(oid);
	}
	
	public X509Downloader(int amount, int threads) throws CertificateException, IOException
	{
		this.threads = threads;
		this.initHashMap();
		this.initiateTopSites(amount);
		this.certificatefactory = CertificateFactory.getInstance("X.509");
		this.socketFactory = (SSLSocketFactory) SSLSocketFactory.getDefault();
	}
	
	public X509Downloader(int amount, int threads, String fileName) throws CertificateException, IOException
	{
		this.threads = threads;
		this.initHashMap();
		this.initiateTopSites(amount);
		this.certificatefactory = CertificateFactory.getInstance("X.509");
		this.socketFactory = (SSLSocketFactory) SSLSocketFactory.getDefault();
		this.fileName = fileName+".txt";
	}
	
	private void initHashMap()
	{
		certTypeMap.put("2.23.140.1.2.1","DV");
		certTypeMap.put("2.23.140.1.2.2", "OV");
		certTypeMap.put("1.3.159.1.17.1","EV");
	    certTypeMap.put("1.3.6.1.4.1.34697.2.1","EV");
	    certTypeMap.put("1.3.6.1.4.1.34697.2.2","EV");
	    certTypeMap.put("1.3.6.1.4.1.34697.2.3","EV");
	    certTypeMap.put("1.3.6.1.4.1.34697.2.4","EV");
	    certTypeMap.put("1.2.40.0.17.1.22","EV");
	    certTypeMap.put("2.16.578.1.26.1.3.3","EV");
	    certTypeMap.put("1.3.6.1.4.1.17326.10.14.2.1.2","EV");
	    certTypeMap.put("1.3.6.1.4.1.17326.10.8.12.1.2","EV");
	    certTypeMap.put("1.3.6.1.4.1.6449.1.2.1.5.1","EV");
	    certTypeMap.put("2.16.840.1.114412.2.1","EV");
	    certTypeMap.put("2.16.840.1.114412.1.3.0.2","EV");
	    certTypeMap.put("2.16.528.1.1001.1.1.1.12.6.1.1.1","EV");
	    certTypeMap.put("2.16.792.3.0.4.1.1.4","EV");
	    certTypeMap.put("2.16.840.1.114028.10.1.2","EV");
	    certTypeMap.put("0.4.0.2042.1.4","EV");
	    certTypeMap.put("0.4.0.2042.1.5","EV");
	    certTypeMap.put("1.3.6.1.4.1.13177.10.1.3.10","EV");
	    certTypeMap.put("1.3.6.1.4.1.14370.1.6","EV");
	    certTypeMap.put("1.3.6.1.4.1.4146.1.1","EV");
	    certTypeMap.put("2.16.840.1.114413.1.7.23.3","EV");
	    certTypeMap.put("1.3.6.1.4.1.14777.6.1.1","EV");
	    certTypeMap.put("2.16.792.1.2.1.1.5.7.1.9","EV");
	    certTypeMap.put("1.3.6.1.4.1.22234.2.5.2.3.1","EV");
	    certTypeMap.put("1.3.6.1.4.1.782.1.2.1.8.1","EV");
	    certTypeMap.put("1.3.6.1.4.1.8024.0.2.100.1.2","EV");
	    certTypeMap.put("1.2.392.200091.100.721.1","EV");
	    certTypeMap.put("2.16.840.1.114414.1.7.23.3","EV");
	    certTypeMap.put("1.3.6.1.4.1.23223.2","EV");
	    certTypeMap.put("1.3.6.1.4.1.23223.1.1.1","EV");
	    certTypeMap.put("2.16.756.1.83.21.0","EV");
	    certTypeMap.put("2.16.756.1.89.1.2.1.1","EV");
	    certTypeMap.put("2.16.840.1.113733.1.7.48.1","EV");
	    certTypeMap.put("2.16.840.1.114404.1.1.2.4.1","EV");
	    certTypeMap.put("2.16.840.1.113733.1.7.23.6","EV");
	    certTypeMap.put("1.3.6.1.4.1.6334.1.100.1","EV");
	    certTypeMap.put("2.16.840.1.114171.500.9","EV");
	    certTypeMap.put("1.3.6.1.4.1.36305.2", "EV");
	}
	
	private void initiateTopSites(int amount) throws CertificateException, IOException
	{
		this.list = new ArrayList<>(amount);
		this.topSites = new LinkedHashMap<String, Boolean>(amount);
		//start with some dummy sites with the amount of threads.
		for(int i = 0; i < this.threads; i++)
		{
			this.topSites.put("fotbollskanalen.se?a="+i, false);
		}
		
		// Open the file
		FileInputStream fstream = new FileInputStream("top-1m.csv");
		BufferedReader br = new BufferedReader(new InputStreamReader(fstream));

		String strLine;
		int count = 0;
		//Read File Line By Line
		while ((strLine = br.readLine()) != null && count < amount)   {
			// Print the content on the console
			String topSite = strLine.split(",")[1];
			this.topSites.put(topSite, false);
			count++;
		}
		
		// end with the same dummy sites
		for(int i = 0; i < this.threads; i++)
		{
			int j = i + this.threads;
			this.topSites.put(this.dummySite + "?a="+j, false);
		}

		//Close the input stream
		br.close();
	}
	
	private String dummySite = "fotbollskanalen.se";
	
	private TLSClient downloadCertificateSynchronized(String site, String cert_path) throws CertificateException, UnknownHostException, IOException, ConnectException 
	{
		TLSClient tc = new TLSClient(site);
		tc.connect();
		TLSClient tc2 = new TLSClient(site);
		tc2.connect();
		
		return tc;
	}
	
	private TLSClient downloadCertificateSynchronizedWithoutSCT(String site, String cert_path) throws CertificateException, UnknownHostException, IOException, ConnectException 
	{
		TLSClient tc = new TLSClient(site, false);
		tc.connect();
		
		return tc;
	}
	
	private long downloadSCTSynchronized(String site, String sct_path) throws IOException, InterruptedException
	{
		String cmd = "openssl s_client -serverinfo 18 -connect "+ site +":443 -servername "+ site +" </dev/null 2>/dev/null";
		String cmd2 = "| sed -n -e \"/BEGIN SERVERINFO/,/END SERVERINFO/p\" > " + sct_path+".pem";
		String[] arga = new String[] {"/bin/sh", "-c", cmd, cmd2};
		long starttime = System.nanoTime();
		Process proc = new ProcessBuilder(arga).start();
		proc.waitFor();
		long endTime = (System.nanoTime() - starttime);
		
		cmd = "grep -v \"\\---\" " + sct_path + ".pem | base64 -D > "+ sct_path +".der";
		arga = new String[] {"/bin/sh", "-c", cmd};
		Process proc2 = new ProcessBuilder(arga).start();
		proc2.waitFor();
		
		return endTime;
	}
	
	private X509Certificate generateX509FromFile(String fileName) throws FileNotFoundException, CertificateException
	{
		FileInputStream fis = new FileInputStream(fileName);
		return (X509Certificate)certificatefactory.generateCertificate(fis);
	}
	
	private SCTExtension generateSCTFromFile(String fileName) throws CertificateException, IOException
	{
		FileInputStream fis = new FileInputStream(fileName);
		return new SCTExtension(IOUtils.toByteArray(fis));
	}
	
	private int count = 0;
	
	private String[] getSite(){
		int count = 0;
		for(Entry<String, Boolean> e : this.topSites.entrySet())
		{
			synchronized(e)
			{
				if(!e.getValue())
				{
					e.setValue(true);
					return new String[]{e.getKey(), String.valueOf(count)};
				}
			}
			count++;
		}
		return null;
	}
	
	private void writeStartUp() throws IOException
	{
		File readme = new File("READ_ME.txt");
		readme.createNewFile();
		PrintWriter writer = new PrintWriter("READ_ME.txt", "UTF-8");
		writer.println("This information is what program has to run with: ");
		writer.println("Available processors (cores): " + Runtime.getRuntime().availableProcessors());
		writer.println("Free memory (bytes): " +  Runtime.getRuntime().freeMemory());
		long maxMemory = Runtime.getRuntime().maxMemory();
	    /* Maximum amount of memory the JVM will attempt to use */
		writer.println("Maximum memory (bytes): " + 
	        (maxMemory == Long.MAX_VALUE ? "no limit" : maxMemory));
		writer.println("Total memory available to JVM (bytes): " + 
		        Runtime.getRuntime().totalMemory());
		writer.println("OS: " + System.getProperty("os.name"));
		//code here will be hard written because it will be hard implement dynamically via Java
		writer.println("CPU: Intel Core i7-4770K CPU @ 3.50GHz");
		writer.close();
	}
	
	public void getCertificates() throws CertificateException, FileNotFoundException, IOException, InterruptedException
	{
		this.writeStartUp();
		File f = new File(this.fileName);
		f.createNewFile();
		PrintWriter writer = new PrintWriter(this.fileName, "UTF-8");
		Decoder decoder = Base64.getDecoder();
		final String delimiter = ";;;";
		final SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
		int start = 0;
		final int threads = this.threads;
		
		for(int i = 0; i < threads; i++)
		{
			final int tStart = start;
			new Thread(new Runnable() {
				
				@Override
				public void run() {
					while(true)
					{
						String[] values = getSite();
						if(values == null)
							break;
						BufferedWriter progWriter = null;
						try {
							progWriter = new BufferedWriter(new FileWriter("progress.txt"));
						} catch (IOException e2) {
							// TODO Auto-generated catch block
							e2.printStackTrace();
						}
						String site = values[0];
						int pos = Integer.parseInt(values[1]);
						try
						{
							for(int k = 0; k < 3; k++)
							{
								long totSiteTimea = System.nanoTime();
								String cert_path = "Certs/"+site+".pem";
								String sct_path = "SCT/"+site;
		
								long starttime = System.nanoTime();
								
								TLSClient tc = null;
								TLSClient tc2 = null;
								SCTExtension sct = null;
								boolean emptyFileX509 = false;
								boolean timedOut = false;
								boolean unknownHost = false;
								boolean tlsError = false;
								boolean unknownError = false;
								boolean failed = false;
								Exception ex = null;
								try
								{ 
									
									tc = downloadCertificateSynchronized(site, cert_path);
									tc2 = downloadCertificateSynchronizedWithoutSCT(site, cert_path);
									
								}
								catch(Exception e)
								{
									if(e instanceof ConnectException)
									{
										if(e.getMessage().contains("Connection timed out: connect") || e.getMessage().contains("Operation timed out"))
											timedOut = true;
										else if(e.getMessage().equals("Connection refused: connect"))
											emptyFileX509 = true;
										else
											unknownError = true;
									}
									else if(e instanceof IOException)
										tlsError = true;
									else if(!(e instanceof UnknownHostException))
										unknownHost = true;
									else
										unknownError = true;
									
									ex = e;
									failed = true;
									System.out.println(e.getMessage());
								}
								boolean problemWithSCT = false;
								SCTExtension scte = null;
								try{
									if(!failed && tc.usedX509v3Extension())
										scte = new SCTExtension(tc.getX509SCT());
									else if(!failed && tc.usedTLSExtension())
										scte = new SCTExtension(tc.getTLSSCT());
									else if(!failed && tc.usedOCSPStapling())
										scte = new SCTExtension(tc.getOCSPSCT());
								}
								catch(Exception e)
								{
									problemWithSCT = true;
								}
								
								boolean valid = true;
								long validationTime = System.nanoTime();
								if(!failed)
								{
									try{tc.getCertificate().checkValidity();}
									catch(Exception e){valid = false;}
								}
								validationTime = System.nanoTime() - validationTime;
		
								if(emptyFileX509)
								{
									synchronized(list){
										list.add(new SimpleEntry<Integer, String>(pos, "(!FAIL 1!) " +  site + " : does not have any x509 certificate or we could not find it"));
									}
									break;
								}
								else if(unknownError)
								{
									String a = ex.getMessage();
									if(a != null)
										a = a.replace("\n", "");
									synchronized(list){
										list.add(new SimpleEntry<Integer, String>(pos, "(!FAIL 2!) " +  site + " : Unknown error. Error message: " + a));
									}
									break;
								}
								else if(timedOut && k == 1)
								{
									synchronized(list){
										list.add(new SimpleEntry<Integer, String>(pos, "(!FAIL 3!) " +  site + " : could not connect to this site, timed out"));
									}
									break;
								}
								else if(unknownHost)
								{
									synchronized(list){
										list.add(new SimpleEntry<Integer, String>(pos, "(!FAIL 4!) " +  site + " : Could not find a website with this address, unknown host"));
									}
									break;
								}
								else if(tlsError)
								{
									synchronized(list){
										list.add(new SimpleEntry<Integer, String>(pos, "(!FAIL 5!) " +  site + " : A TLS error happend, could be an attack, TLS ERROR"));
									}
									break;
								}
								else if(!failed)
								{
									DateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
									Date date = new Date();
									long totSiteTime =  System.nanoTime() - totSiteTimea;
									java.security.cert.X509Certificate x509 = tc.getCertificate();
									String certType = getCertificateType(getCertificatePolicyId(x509));
									StringBuilder sb = new StringBuilder();
									
									sb.append(x509.getSerialNumber() + delimiter + x509.getVersion()+ delimiter + x509.getType());
									sb.append(delimiter + x509.getSigAlgName() + delimiter + x509.getSigAlgOID()
										+ delimiter + x509.getIssuerX500Principal().getName());
									sb.append(delimiter + boolArrToString(x509.getKeyUsage()) + delimiter + x509.getSignature().length + delimiter + x509.getPublicKey().getEncoded().length);
									sb.append(delimiter + sdf.format(x509.getNotBefore()) + delimiter + sdf.format(x509.getNotAfter()));
									sb.append(delimiter + x509.getSubjectDN().getName() + delimiter + x509.getSubjectUniqueID());
									sb.append(delimiter + tc.getHandshakeTime() + delimiter + tc.getOCSPTime() + delimiter + tc.getTLSTime());
									sb.append(delimiter + tc.getX509Time() + delimiter + tc.usedOCSPStapling() + delimiter + tc.usedTLSExtension());
									sb.append(delimiter + tc.usedX509v3Extension() + delimiter + tc.getAmountBytesReceived() + delimiter + tc.getAmountBytesSent());
									sb.append(delimiter + tc.getOCSPSCT().length + delimiter + tc.getTLSSCT().length + delimiter + tc.getX509SCT().length);
									sb.append(delimiter + DatatypeConverter.printBase64Binary(x509.getPublicKey().getEncoded()));
									sb.append(delimiter + x509.getPublicKey().getAlgorithm() + delimiter + validationTime + delimiter + valid);
									sb.append(delimiter + site + delimiter + pos + delimiter + tc.getIpAddress() + delimiter + dateFormat.format(date));
									sb.append(delimiter + tc2.getHandshakeTime() + delimiter + certType + delimiter);
									if(!problemWithSCT && scte != null)
									{
										sb.append(scte.getDecodeTime() + delimiter + scte.getLenght() + delimiter + scte.getScts().size() + delimiter);
										for(SCTExtension.SCT sct1 : scte.getScts())
										{
											sb.append("|||");
											sb.append(sct1.GetLog() + delimiter + sct1.getLogId());
											sb.append(delimiter + sct1.getTimestampDateString() + delimiter + sct1.getVersion());
											sb.append(delimiter + sct1.getLenght() + delimiter + sct1.getHashAlgorithm() + delimiter + sct1.getSignatureAlgorithm());
											sb.append(delimiter + sct1.getSignatureLen() + delimiter + sct1.getSignature());
											sb.append("|||");
										}
									}
									synchronized(list){
										list.add(new SimpleEntry<Integer, String>(pos, sb.toString().replace("\n", "")));
									}
									break;
								}
								else if(!timedOut)
								{
									String a = "";
									if(ex != null)
									{
										a = ex.getMessage();
										if(a != null)
											a = a.replace("\n", "");
									}
									synchronized(list){
										list.add(new SimpleEntry<Integer, String>(pos, "(!FAIL 6!) " +  site + " : No idea what happend here! error message: " + a));
									}
									break;
								}
							}
						}
						catch(Exception e)
						{
							e.printStackTrace();
							String a = e.getMessage();
							if(a != null)
								a = a.replace("\n", "");
							synchronized(list){
								list.add(new SimpleEntry<Integer, String>(pos, "(!FAIL 7!) " +  site + " : something unexpected happend, error - " + a));
							}
						}						
						
						String print = list.size() + " / " + topSites.size() + " - " + site + " website number " + pos;
						System.out.println(print);
					}
				}
			}).start();
			Thread.sleep(100);
			start++;
		}
		
		
		while(this.list.size() < this.topSites.size())
		{
			synchronized (this.list) {
				if(this.list.size() >= this.topSites.size())
					break;
			}
		}
		list.sort(new Comparator<SimpleEntry<Integer, String>>() {

			@Override
			public int compare(SimpleEntry<Integer, String> o1, SimpleEntry<Integer, String> o2) {
				// TODO Auto-generated method stub
				return o1.getKey().compareTo(o2.getKey());
			}
			
		});
		for(int i = 0 ; i < this.threads; i++)
		{
			list.remove(this.list.size() - 1);
		}
		for(int i = 0; i < this.threads; i++)
		{
			list.remove(0);
		}
		for(SimpleEntry<Integer, String> se : list)
		{
			writer.println(se.getValue());
		}
		writer.close();
		System.out.println("done!");
	}
	
	private String boolArrToString(boolean[] arr)
	{
		StringBuilder sb = new StringBuilder();
		if(arr == null)
			return "no key usage found";
		for(boolean b : arr)
		{
			if(b)
				sb.append("true,");
			else
				sb.append("false,");
		}
		return sb.toString().substring(0, sb.length()-1);
	}

	private static final String CERTIFICATE_POLICY_OID = "2.5.29.32";
	
	public static HashSet<String> getCertificatePolicyId(X509Certificate cert) throws IOException {
        HashSet<String> hs = new HashSet<>();
        byte[] extPolicyBytes = cert.getExtensionValue(CERTIFICATE_POLICY_OID);
        if (extPolicyBytes == null) {
            return hs;
        }
 
        DEROctetString oct = (DEROctetString) (new ASN1InputStream(new ByteArrayInputStream(extPolicyBytes)).readObject());
        ASN1Sequence seq = (ASN1Sequence) new ASN1InputStream(new ByteArrayInputStream(oct.getOctets())).readObject();
        for(int i = 0; i < seq.size(); i++)
        {
        	CertificatePolicies certificatePolicies = new CertificatePolicies(PolicyInformation.getInstance(seq.getObjectAt(i)));
	        PolicyInformation[] policyInformation = certificatePolicies.getPolicyInformation();
	        for(int j = 0; j < policyInformation.length; j++)
	        {
	            hs.add(policyInformation[j].getPolicyIdentifier().getId());
	        }  
        }
       return hs;
    }
	
	public String getCertificateType(HashSet<String> oids)
    {
		for(String oid : oids)
		{
	        if(certTypeMap.containsKey(oid))
	        {
	        	String returnType;
	            if(certTypeMap.get(oid).equals("EV"))
	            	returnType = "EV";
	            else if(certTypeMap.get(oid).equals("DV"))
	            	returnType = "DV";
	            else if(certTypeMap.get(oid).equals("OV"))
	            	returnType = "OV";
	            else
	            	returnType = "Other";
	            
	            return returnType;
	        }
		}
        return "Other";
    }

	/**
	 * @return the dummySite
	 */
	public String getDummySite() {
		return dummySite;
	}

	/**
	 * @param dummySite the dummySite to set
	 */
	public void setDummySite(String dummySite) {
		this.dummySite = dummySite;
	}

}
